from Supplychain.Generic.cosmo_api_parameters import CosmoAPIParameters
from Supplychain.Generic.folder_io import FolderWriter, FolderReader
from Supplychain.Wrappers.environment_variables import SECONDS_IN_MINUTE
from Supplychain.Generic.excel_folder_reader import ExcelReader
from Supplychain.Generic.csv_folder_reader import CSVReader
from Supplychain.Generic.memory_folder_io import MemoryFolderIO
from Supplychain.Transform.from_table_to_dict import write_transformed_data
from Supplychain.Generic.timer import Timer
from dateutil import parser
import math
from typing import Union

GRANULARITIES = {
    "minute": 1,
    "hour": 60,
    "day": 60 * 24,
    "week": 60 * 24 * 7,
    "month": 60 * 24 * 30,
    "quarter": 60 * 24 * 90,
    "year": 60 * 24 * 365,
}


def object_to_bool(item: Union[object, str]) -> bool:
    if type(item) is str:
        return item.lower() == "true"
    return bool(item)


class DictPatcher(Timer):
    def __write_updated_files(self):
        to_use_reader = self.memory

        # ADT column names to be replaced for post ADT simulation
        name_replacements = {"$sourceId": "source", "$targetId": "target", "$id": "id"}

        for file_name, file_content in to_use_reader.files.items():
            self.writer.write_from_list(
                dict_list=[
                    {
                        name_replacements.get(item_k, item_k): item_v
                        for item_k, item_v in item.items()
                    }
                    for item in file_content
                ],
                file_name=file_name,
            )

    def __init__(
        self, reader: FolderReader, writer: FolderWriter, parameters: CosmoAPIParameters
    ):
        Timer.__init__(self, "[ParameterHandler]")
        self.parameters = parameters
        self.writer = writer
        self.reader = reader

        self.memory = MemoryFolderIO()
        for file_name, file_content in self.reader.files.items():
            self.memory.write_from_list(dict_list=file_content, file_name=file_name)
        if "Configuration" not in self.memory.files:
            self.memory.files["Configuration"] = [
                dict(),
            ]

    def handle_mass_action_lever(self):
        try:
            lever_folder = self.parameters.get_dataset_path("mass_lever_excel_file")
        except ValueError:
            self.display_message("No mass action lever found - skipping")
            return False, 0
        lever_reader = ExcelReader(lever_folder, keep_nones=False)
        self.memory.reset()

        reading_errors = write_transformed_data(reader=lever_reader, writer=self.memory)
        if reading_errors:
            return True, reading_errors
        if not self.memory.files.setdefault("Configuration", []):
            self.memory.files["Configuration"].append({})

        self.__write_updated_files()
        return True, 0

    def handle_simple_simulation(self):
        configuration = self.memory.files["Configuration"][0]

        start_date = parser.isoparse(
            self.parameters.get_named_parameter("start_date").value
        )
        end_date = parser.isoparse(
            self.parameters.get_named_parameter("end_date").value
        )
        simulation_granularity = GRANULARITIES.get(
            self.parameters.get_named_parameter("simulation_granularity").value,
            GRANULARITIES["day"],
        )

        duration = (end_date - start_date).total_seconds() // SECONDS_IN_MINUTE

        steps = int(math.ceil(duration / simulation_granularity))

        self.display_message(f"Starting Date: {start_date.isoformat()}")
        self.display_message(f"Simulated Cycles: {steps}")
        self.display_message(f"Steps per Cycle: {1}")
        self.display_message(f"TimeStep Duration: {simulation_granularity} minutes")

        configuration["StartingDate"] = start_date.isoformat()
        configuration["SimulatedCycles"] = steps
        configuration["StepsPerCycle"] = 1
        configuration["TimeStepDuration"] = simulation_granularity

        self.__handle_demand_plan()
        self.__handle_transport_duration()
        self.__handle_production_resource_opening_time()

        self.__write_updated_files()

    def handle_optimization_parameter(self):
        parameters = (('optimization_objective', 'OptimizationObjective', 'Optimization Objective', None),)
        for name, config_name, display_name, type_f in parameters:
            self.__handle_configuration_parameter(name, config_name, display_name, type_f)
        self.__write_updated_files()

    def handle_flow_management_policies(self):
        parameters = (
            ("stock_policy", "Stock", "StockPolicy"),
            ("sourcing_policy", "Stock", "SourcingPolicy"),
            ("stock_dispatch_policy", "Stock", "DispatchPolicy"),
            ("production_policy", "ProductionResource", "ProductionPolicy"),
        )
        for parameter_name, entity_type, entity_parameter_name in parameters:
            try:
                parameter_value = self.parameters.get_named_parameter(
                    parameter_name
                ).value
            except ValueError:
                self.display_message(f"{parameter_name} is not defined - skipping")
                continue
            if parameter_value == "FromDataset":
                continue
            if entity_type not in self.memory.files:
                continue
            entities = self.memory.files[entity_type]
            for entity in entities:
                entity[entity_parameter_name] = parameter_value
            entities_count = len(entities)
            self.display_message(
                f"Set {entity_parameter_name} to {parameter_value} for all {entity_type} "
                f'({entities_count} entit{"ies" if entities_count > 1 else "y"})'
            )
        self.__write_updated_files()

    def handle_model_behavior(self):
        parameters = (
            (
                "manage_backlog_quantities",
                "ManageBacklogQuantities",
                "Manage Backlog Quantities",
                object_to_bool,
            ),
            (
                "empty_obsolete_stocks",
                "EmptyObsoleteStocks",
                "Empty Obsolete Stocks",
                object_to_bool,
            ),
            ("batch_size", "BatchSize", "Batch Size", int),
            (
                "financial_cost_of_stocks",
                "FinancialCostOfStock",
                "Financial Cost Of Stock",
                float,
            ),
            ("carbon_tax", "CarbonTax", "Carbon Tax", float),
            (
                "intermediary_stock_dispatch",
                "IntermediaryStockDispatchPolicy",
                "Intermediary Stock Dispatch Policy",
                None,
            ),
            (
                "actualize_shipments",
                "ActualizeShipments",
                "Actualize Shipments",
                object_to_bool,
            ),
        )
        for name, config_name, display_name, type_f in parameters:
            self.__handle_configuration_parameter(
                name, config_name, display_name, type_f
            )
        self.__write_updated_files()

    def handle_uncertainties_settings(self):
        parameters = (
            (
                "uncertainties_probability_distribution",
                "UncertaintiesProbabilityDistribution",
                "Uncertainties Probability Distribution",
                None,
            ),
            (
                "transport_uncertainty_distribution",
                "TransportUncertaintiesProbabilityDistribution",
                "Transport Uncertainties Probability Distribution",
                None,
            ),
            (
                "sample_size_uncertainty_analysis",
                "SampleSizeUncertaintyAnalysis",
                "Sample Size Uncertainty Analysis",
                int,
            ),
        )
        for name, config_name, display_name, type_f in parameters:
            self.__handle_configuration_parameter(
                name, config_name, display_name, type_f
            )
        self.__write_updated_files()

    def __handle_configuration_parameter(
        self,
        parameter_name: str,
        configuration_parameter_name: str,
        display_name: str,
        type_function,
    ) -> bool:
        configuration = self.memory.files["Configuration"][0]
        try:
            parameter_value = self.parameters.get_named_parameter(parameter_name).value
        except ValueError:
            self.display_message(f"{parameter_name} is not defined - skipping")
            return False
        param_value = (
            parameter_value if type_function is None else type_function(parameter_value)
        )
        self.display_message(f"{display_name}: {param_value}")
        configuration[configuration_parameter_name] = param_value
        return True

    def __handle_demand_plan(self):
        try:
            demand_plan_folder = self.parameters.get_dataset_path("demand_plan")
        except ValueError:
            self.display_message("No demand plan found - skipping")
            return
        reader = CSVReader(demand_plan_folder)
        stock_changes = dict()
        for k, demand_plan in reader.files.items():
            for _demand in demand_plan:
                stock_id = _demand.get("id")
                timestep = str(_demand.get("Timestep"))
                stock_changes.setdefault(stock_id, dict())
                for key in ["Demands", "DemandUncertainties", "DemandWeights"]:
                    stock_changes[stock_id].setdefault(key, dict())
                    if key in _demand:
                        stock_changes[stock_id][key][timestep] = _demand.get(key)

        stocks = self.memory.files["Stock"]
        for stock in stocks:
            _id = stock["id"]
            for key in ["Demands", "DemandUncertainties", "DemandWeights"]:
                if _id in stock_changes:
                    stock[key] = stock_changes[_id][key]
                else:
                    stock[key] = {}

    def __handle_transport_duration(self):
        try:
            transport_duration_folder = self.parameters.get_dataset_path(
                "transport_duration"
            )
        except ValueError:
            self.display_message("No lever for transport duration found - skipping")
            return
        reader = CSVReader(transport_duration_folder)

        duration_changes = dict()
        for k, transport_durations in reader.files.items():
            for _td in transport_durations:
                source = _td.get("source")
                target = _td.get("target")
                duration = int(_td.get("Duration"))
                duration_changes[(source, target)] = duration
        for e in self.memory.files["Transport"]:
            duration = duration_changes.get((e.get("source"), e.get("target")))
            if duration:
                e["Duration"] = duration

    def __handle_production_resource_opening_time(self):
        try:
            pr_opening_time_folder = self.parameters.get_dataset_path(
                "production_resource_opening_time"
            )
        except ValueError:
            self.display_message(
                "No lever for production resource opening time found - skipping"
            )
            return
        reader = CSVReader(pr_opening_time_folder)

        opening_changes = dict()
        for k, changes in reader.files.items():
            for _td in changes:
                _id = _td.get("id")
                timestep = _td.get("Timestep")
                opening_time = float(_td.get("OpeningTimes"))
                opening_changes.setdefault(_id, dict())
                opening_changes[_id][timestep] = opening_time
        for e in self.memory.files.get("ProductionResource", []):
            change = opening_changes.get(e["id"])
            if change:
                e["OpeningTimes"] = change

    def handle_stochastic_optimization_parameters(self, param_to_update: dict):
        parameters = (
            (
                "KPI",
                "Kpi",
                "KPI",
                None,
            ),
            (
                "optimization_mode",
                "OptimizationMode",
                "Optimization Mode",
                None,
            ),
            (
                "stats",
                "Statistic",
                "Statistic",
                None,
            ),
            (
                "target_value",
                "TargetedValue",
                "Targeted Value",
                float,
            ),
            (
                "decision_variable",
                "DecisionVariable",
                "Decision Variable",
                None,
            ),
            (
                "decision_variable_min",
                "DecisionVariableMin",
                "DecisionVariable Min",
                float,
            ),
            (
                "decision_variable_max",
                "DecisionVariableMax",
                "DecisionVariable Max",
                float,
            ),
            (
                "max_duration_of_optim",
                "OptimizationMaximalDuration",
                "Optimization Maximal Duration",
                float,
            ),
            (
                "optimization_algorithm",
                "OptimizationAlgorithm",
                "Optimization Algorithm",
                None,
            ),
            (
                "optim_batch_size",
                "OptimizationBatchSize",
                "Optimization Batch Size",
                int,
            ),
            (
                "uncertainties_probability_distribution",
                "UncertaintiesProbabilityDistribution",
                "Uncertainties Probability Distribution",
                None,
            ),
            (
                "transport_uncertainty_distribution",
                "TransportUncertaintiesProbabilityDistribution",
                "Transport Uncertainties Probability Distribution",
                None,
            ),
            (
                "sample_size_uncertainty_analysis_optim",
                "SampleSizeUncertaintyAnalysis",
                "Sample Size Uncertainty Analysis",
                int,
            ),
            (
                "final_sample_size_uncertainty_analysis",
                "FinalSampleSizeUncertaintyAnalysis",
                "Sample Size Final Uncertainty Analysis",
                int,
            ),
            (
                "max_iterations_for_optim",
                "MaxIterationsForOptim",
                "Max Iterations For Optim",
                int,
            ),
            (
                "automatic_parallelization_config",
                "AutomaticParallelizationConfig",
                "Automatic Parallelization Config",
                object_to_bool,
            ),
            (
                "max_number_of_sim_in_parallel",
                "MaxNumberOfSimInParallel",
                "Max Number Of Sim In Parallel",
                int,
            ),
        )

        for name, config_name, display_name, type_f in parameters:
            self.__handle_configuration_parameter(
                name, config_name, display_name, type_f
            )
        if param_to_update is not None:
            for key, value in param_to_update.items():
                configuration = self.memory.files["Configuration"][0]
                configuration[key] = value

        self.__write_updated_files()



import multiprocessing
from typing import Union
from copy import deepcopy

import pandas
import comets as co

from Supplychain.Wrappers.simulator import CosmoEngine
from Supplychain.Generic.adx_and_file_writer import ADXAndFileWriter
from Supplychain.Generic.timer import Timer
from Supplychain.Run.simulation import run_simple_simulation
from Supplychain.Run.consumers import (
    StockConsumer,
    PerformanceConsumer,
    StocksAtEndOfSimulationConsumer,
)
from Supplychain.Run.uncertainty_analysis_helper_functions import (
    get_max_time_step,
    get_attribute,
    get_stocks,
    get_transports,
    collect_transport_information,
    extend_dic,
    add_tag_to_parameterset,
    collect_simulated_stock_final_output,
    collect_simulated_stock_output,
    collect_simulated_transport_output,
    transform_data,
    create_transport_distribution_sampling,
    create_demand_generator,
)


class UncertaintyAnalyzer:
    """
    Object in charge of performing the different steps of the uncertainty analysis.
    Its main method is "execute".

    Args:
        simulation_name (str): Name of simulation, used by the probes
        simulation_path (str): Name of the simulation file (typically Simulation)
        sample_size (int): Number of simulations runs by the uncertainty analysis
        batch_size (int): Number of simulations runs that are run in a same batch by the uncertainty analysis
        amqp_consumer_adress (Union[str, None], optional): Adress of consumer to send probe results to. Defaults to None.
        consumers (list, optional): Which consumers are activated. Defaults to ["Stocks", "Transport", "Performances", "StocksAtEndOfSimulation", "PerformanceAMQP"].
        validation_folder (str, optional): Local folder to which results are written to, used by the tests. Defaults to None.
        cold_inputs (dict, optional): Parameters that are passed to the simulator at each simulation and don't change during the analysis. Defaults to {}.
        timer (Timer object, optional): Timer object that can be used for logs and counting time. Defaults to None.
    """

    def __init__(
        self,
        simulation_name,
        simulation_path,
        sample_size,
        batch_size,
        amqp_consumer_adress=None,
        consumers=["Stocks", "Transport", "Performances", "StocksAtEndOfSimulation"],
        validation_folder=None,
        cold_inputs={},
        timer=None,
        n_jobs=-1,
    ):
        if timer is None:
            self.t = Timer("[Run Uncertainty]")
        else:
            self.t = timer
        self.simulation_name = simulation_name
        self.simulation_path = simulation_path
        self.amqp_consumer_adress = amqp_consumer_adress
        self.sample_size = sample_size
        self.batch_size = batch_size
        self.validation_folder = validation_folder
        self.consumers = consumers
        self.cold_inputs = cold_inputs

        if self.batch_size > self.sample_size:
            self.batch_size = self.sample_size

        if n_jobs == -1:
            self.processes_size = min(multiprocessing.cpu_count(), self.batch_size)
        else:
            self.processes_size = n_jobs

    def execute(self):
        """Setup and run the uncertainty analysis

        Returns:
            dict: dictionary with keys among ["Stock", "Transport", "Performances", "StocksAtEndOfSimulation"]
                containing the output tables. Which table is available depends on the specified
                consumers of the UncertaintyAnalyzer.
        """
        self.t.split("Initialize uncertainty analysis")
        self.initialize_simulator_interface()
        self.collect_simulation_parameters()
        self.create_encoder()
        self.create_get_outcomes()
        self.create_sampling()
        self.create_task(self.cold_inputs)
        self.t.split(
            "Ended uncertainty analysis initialization : {time_since_last_split}"
        )

        self.t.display_message("Run uncertainty analysis")
        self.run_experiment()
        self.t.split("Ended uncertainty analysis run : {time_since_last_split}")

        self.t.display_message("Reformat uncertainty analysis results")
        self.reformat_results()
        if self.validation_folder:
            self.write_results_locally()
        self.t.split(
            "Ended uncertainty analysis reformatting : {time_since_last_split}"
        )

        return self.results

    def initialize_simulator_interface(self):
        used_probes = []
        used_consumers = []
        custom_consumers = []

        if "Stocks" in self.consumers:
            used_probes.append("Stocks")
            custom_consumers.append((StockConsumer, "LocalConsumer", "Stocks"))
        if "StocksAtEndOfSimulation" in self.consumers:
            used_probes.append("StocksAtEndOfSimulation")
            custom_consumers.append(
                (StocksAtEndOfSimulationConsumer, "StocksAtEndOfSimulationConsumer", "StocksAtEndOfSimulation")
            )
        if "Performances" in self.consumers:
            used_probes.append("PerformanceIndicators")
            custom_consumers.append(
                (
                    PerformanceConsumer,
                    "LocalPerformanceConsumer",
                    "PerformanceIndicators",
                )
            )
        if "PerformanceAMQP" in self.consumers:
            used_consumers.append("PerformanceIndicatorsAMQP")

        self.simulator_interface = co.CosmoInterface(
            simulator_path=self.simulation_path,
            custom_sim_engine=CosmoEngine,
            simulation_name=self.simulation_name,
            amqp_consumer_address=self.amqp_consumer_adress,
            used_consumers=used_consumers,
            used_probes=used_probes,
            custom_consumers=custom_consumers,
        )

    def collect_simulation_parameters(self):
        """
        Collect values of attributes of the model
        Used to retrieve the probability distribution parameters, the theoretical transport duration, etc...
        """
        # Load simulator to be able to access attributes of the model
        self.simulator_interface.initialize()

        # Retrieving model information
        self.max_time_step = get_max_time_step(self.simulator_interface)
        self.ActivateCorrelatedDemandUncertainties = get_attribute(
            self.simulator_interface, "Model::@ActivateCorrelatedDemandUncertainties"
        )
        self.DemandCorrelations = get_attribute(
            self.simulator_interface, "Model::@DemandCorrelations"
        )
        self.transport_distribution = get_attribute(
            self.simulator_interface,
            "Model::@TransportUncertaintiesProbabilityDistribution",
        )
        self.ActivateUncertainties = get_attribute(
            self.simulator_interface,
            "Model::@ActivateUncertainties",
        )

        # Getting the name of all the stocks with uncertain demand
        if self.ActivateCorrelatedDemandUncertainties:
            self.uncertain_stocks = get_stocks(self.simulator_interface)
        else:
            self.uncertain_stocks = []

        # Getting the name of all the transport operations
        self.all_transports = get_transports(self.simulator_interface)

        # Collect information about transports:
        # which transport have uncertain durations, what are the schedule for the theoretical transport duration
        # and for the parameters of the probability distribution
        (
            self.transports_with_uncertain_duration,
            self.actual_duration_schedule,
            uncertainty_param_1,
            uncertainty_param_2,
            uncertainty_param_3,
            uncertainty_param_4,
        ) = collect_transport_information(self.simulator_interface, self.all_transports)

        # Collect model information about demands of each stock
        demands = {}
        for stock in self.uncertain_stocks:
            demands[stock] = get_attribute(
                self.simulator_interface,
                f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{stock}::@Demand",
            )

        self.simulator_interface.terminate()
        self.transport_distribution_params = co.DistributionRegistry.information[
            str(self.transport_distribution)
        ]["parameters"]

        # Extending the dictionaries above for scheduled attributes so that all time steps are present
        self.extended_actual_duration = extend_dic(
            self.actual_duration_schedule, self.max_time_step
        )
        self.extended_param_1 = extend_dic(uncertainty_param_1, self.max_time_step)
        self.extended_param_2 = extend_dic(uncertainty_param_2, self.max_time_step)
        self.extended_param_3 = extend_dic(uncertainty_param_3, self.max_time_step)
        self.extended_param_4 = extend_dic(uncertainty_param_4, self.max_time_step)
        self.extended_demands = extend_dic(demands, self.max_time_step)

        if not self.ActivateUncertainties:
            # Remove uncertain transports and stocks if ActivateUncertainties is false
            self.transports_with_uncertain_duration = []
            self.uncertain_stocks = []

    def create_encoder(self):
        """Create encoder of the task"""

        def encoder(
            parameters,
            extended_demands=self.extended_demands,
            transports_with_uncertain_duration=self.transports_with_uncertain_duration,
            extended_actual_duration=self.extended_actual_duration,
            uncertain_stocks=self.uncertain_stocks,
            max_time_step=self.max_time_step,
        ):
            """
            The encoder takes a parameterset containing the input parameters that are changing at each simulation.
            It has the following format :
            {'{Model} [..] Seed': 3251851028, 'T_@_0': 1.7504164949122698, 'T_@_1': 3.05256524351985, [...],
            'T_@_10': 3.2371899035525793, 'StockA':[1.1, 2., ..., 3.4]}
            The parameters correspond either to:
            - the Seed datapath and its value
            - transport_name +_@_ + time_step, and the duration of this transport at this time step
            - stock_name, and a list of demands for this stock
            It returns a parameterset where the keys are datapaths in the model.
            The values for each transport's duration is transformed in one dictionary {TimeStep: value} and updated as follows:
            new_transport_duration = max(0, round(old_transport_duration +
                sample input drawn from the distribution))
            The values for each stock's demand is set to its ExternalDemand in the attribute Demand, which is a dictionary
            {TimeStep: Composite attribute}.
            {'Model [..] T::@ActualDurationSchedule': {'0': 5, '1': 6, [..] '10': 44},
                '{Model}Model::{Attribute}Seed': 3209521878, ...}
            """
            encoded_parameterset = {}
            if "{Model}Model::{Attribute}Seed" in parameters:
                encoded_parameterset = {
                    "{Model}Model::{Attribute}Seed": parameters[
                        "{Model}Model::{Attribute}Seed"
                    ],
                }
            for transports in transports_with_uncertain_duration:
                ActualDurationSchedule = {}
                for i in range(len(extended_actual_duration[transports])):
                    ActualDurationSchedule[str(i)] = max(
                        0,
                        round(
                            extended_actual_duration[transports][i]
                            + parameters[f"{transports}_@_{i}"]
                        ),
                    )
                encoded_parameterset[
                    f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transports}::@ActualDurationSchedule"
                ] = ActualDurationSchedule
            for stock in uncertain_stocks:
                sample_demand = parameters[f"{stock}"]
                demand_attribute = deepcopy(extended_demands[f"{stock}"])
                for i in range(max_time_step):
                    demand_attribute[i]["ExternalDemand"] = sample_demand[i]
                encoded_parameterset[
                    f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{stock}::@Demand"
                ] = demand_attribute
            return encoded_parameterset

        self.encoder = encoder

    def create_get_outcomes(self):
        """Create the get_outcomes function of the task"""

        def get_outcomes(
            modelinterface,
            consumers=self.consumers,
            all_transports=self.all_transports,
            max_time_step=self.max_time_step,
        ):
            """
            Returns a parameter set with all the model's output. More precisely, the parameter set is the
            the result of the concatenation of up to four parameter sets, depending on the consumers that have been chosen.
            In front of the name of each parameter, we add a '1_', '2_', '3_' or '4_' to identify the 4 original parametersets.
            The first one looks like this:
            {'1_U__&@&__0': 5, '1_U__&@&__1': 5, [...], '1_U__&@&__10': 6}.
            The keys correspond to transport_name + __&@&__ + time_step,
            and the value to the duration of the transport at this time step.
            The second parameter set looks like this:
            {'2_A__&@&__ServedQuantity__&@&__0': 0.0, '2_A__&@&__UnservedQuantity__&@&__0': 0.0}.
            The keys correspond to stock + __&@&__ + category (Demand, ServedQuantity,...)  + __&@&__ +  time_step
            The third parameter set contains the performance indicators {'3_OPEX': 0.0, '3_Profit': 1.0, ...}.
            The fourth parameter set looks like:
            {'4_A__&@&__TotalDemand': 1.0}.
            The keys correspond to stock name + __&@&__ + category (TotalDemand, TotalServedQuantity, ServiceLevelSatisfaction, CycleServiceLevel)
            """

            output_parameterset = {}
            if "Transports" in consumers:
                output_parameterset.update(
                    add_tag_to_parameterset(
                        "1_",
                        collect_simulated_transport_output(
                            all_transports, modelinterface, max_time_step
                        ),
                    )
                )
            if "Stocks" in consumers:
                output_parameterset.update(
                    add_tag_to_parameterset(
                        "2_",
                        collect_simulated_stock_output(
                            modelinterface.LocalConsumer.memory
                        ),
                    )
                )
            if "Performances" in consumers:
                output_parameterset.update(
                    add_tag_to_parameterset(
                        "3_", modelinterface.LocalPerformanceConsumer.memory[0]
                    )
                )
            if "StocksAtEndOfSimulation" in consumers:
                output_parameterset.update(
                    add_tag_to_parameterset(
                        "4_",
                        collect_simulated_stock_final_output(
                            modelinterface.StocksAtEndOfSimulationConsumer.memory
                        ),
                    )
                )
            return output_parameterset

        self.get_outcomes = get_outcomes

    def create_sampling(self):
        """Create the sampling of the uncertainty analysis"""
        # Creating the sampling on transport durations
        self.sampling = create_transport_distribution_sampling(
            self.transport_distribution,
            self.extended_param_1,
            self.extended_param_2,
            self.extended_param_3,
            self.extended_param_4,
            self.transports_with_uncertain_duration,
            self.transport_distribution_params,
            self.max_time_step,
        )
        # Add the seed as an uncertain parameter
        self.sampling.append(
            {
                "name": "{Model}Model::{Attribute}Seed",
                "sampling": "seed_generator",
            }
        )

        # Add the generator of samples on the demand
        if self.ActivateCorrelatedDemandUncertainties:
            self.sampling += create_demand_generator(
                self.extended_demands, self.max_time_step, self.DemandCorrelations
            )

    def create_task(self, cold_inputs={}):
        """Create the task on which the uncertainty analysis will be performed

        Args:
            cold_inputs (dict): ParameterSet containing parameters of the simulator
                that will be applied to each evaluation of the task.
                Allows to modify other attributes than those modified by the uncertainty analysis.
        """
        if (
            self.ActivateCorrelatedDemandUncertainties
        ):  # Correlated demands are not compatible with demands drawn inside the model
            cold_input_parameter_set = {
                "{Model}Model::{Attribute}ActivateUncertainties": 0
            }
        else:
            cold_input_parameter_set = {}

        cold_input_parameter_set.update(cold_inputs)

        self.simulationtask = co.ModelTask(
            modelinterface=self.simulator_interface,
            encode=self.encoder,
            get_outcomes=self.get_outcomes,
            cold_input_parameter_set=cold_input_parameter_set,
        )

    def run_experiment(self):
        """Create and run the uncertainty analysis experiment"""
        if self.validation_folder is not None:
            save_task_history = True
        else:
            save_task_history = False

        self.experiment = co.UncertaintyAnalysis(
            task=self.simulationtask,
            sampling=self.sampling,
            stop_criteria={"max_evaluations": self.sample_size},
            analyzer=["standard", "quantiles"],
            n_jobs=self.processes_size,
            save_task_history=save_task_history,
        )

        self.experiment.run()

    def reformat_results(self):
        """Reformat results of the experiment so that they are compatible with the output tables"""
        self.results = {}
        # Separating the results data on the different types of outputs (stock, transport, performances, stocksatendofsimulation)
        self.experiment.results["statistics"].reset_index(inplace=True)
        self.experiment.results["statistics"]["OutputType"] = (
            self.experiment.results["statistics"]["index"]
            .str.split(pat="_", expand=False, n=1)
            .str[0]
        )
        self.experiment.results["statistics"]["index"] = (
            self.experiment.results["statistics"]["index"]
            .str.split(pat="_", expand=False, n=1)
            .str[1]
        )
        if "Transports" in self.consumers:
            df_transport_duration = self.experiment.results["statistics"][
                self.experiment.results["statistics"]["OutputType"] == "1"
            ]
            df_transport_duration = df_transport_duration.drop("OutputType", axis=1)
            df_transport_final = transform_data(df_transport_duration)
            self.results["Transport"] = df_transport_final
        if "Stocks" in self.consumers:
            df_probe_data = self.experiment.results["statistics"][
                self.experiment.results["statistics"]["OutputType"] == "2"
            ]
            df_probe_data = df_probe_data.drop("OutputType", axis=1)
            df_stock_final = transform_data(df_probe_data)
            self.results["Stock"] = df_stock_final
        if "Performances" in self.consumers:
            performances = self.experiment.results["statistics"][
                self.experiment.results["statistics"]["OutputType"] == "3"
            ]
            performances = performances.drop("OutputType", axis=1)
            performances = performances.rename(columns={"index": "KPI"})
            self.results["Performances"] = performances
        if "StocksAtEndOfSimulation" in self.consumers:
            df_stocksatendofsimulationconsumer = self.experiment.results["statistics"][
                self.experiment.results["statistics"]["OutputType"] == "4"
            ]
            df_stocksatendofsimulationconsumer = df_stocksatendofsimulationconsumer.drop("OutputType", axis=1)
            df_stocksatendofsimulationconsumer = transform_data(df_stocksatendofsimulationconsumer, timestep=False)
            self.results["StocksAtEndOfSimulation"] = df_stocksatendofsimulationconsumer

    def write_results_locally(self):
        """Write the results tables locally to csv files"""

        # Get all demands directly from the experiment, before aggregation of statistics
        demands = []
        j = 0
        for i in self.experiment.task_history["outputs"]:

            for (k, v) in i.items():
                if "__&@&__Demand__&@&__" in k:
                    demand_result_dict = {}
                    demand_result_dict["Simulation"] = j
                    demand_result_dict["Entity"] = k.split("__&@&__Demand__&@&__")[
                        0
                    ].split("_", 1)[1]
                    demand_result_dict["TimeStep"] = k.split("__&@&__Demand__&@&__")[1]
                    demand_result_dict["Demand"] = v

                    demands.append(demand_result_dict)
            j += 1
        demand_df = pandas.DataFrame(demands)
        demand_df.to_csv(
            str(self.validation_folder) + "/df_all_demands.csv", index=False
        )
        self.results["Stock"].to_csv(
            str(self.validation_folder) + "/final_df_comets.csv", index=False
        )
        self.results["StocksAtEndOfSimulation"].to_csv(
            str(self.validation_folder) + "/df_stocksatendofsimulation.csv", index=False
        )
        self.results["Transport"].to_csv(
            str(self.validation_folder) + "/df_transport.csv", index=False
        )
        self.results["Performances"].to_csv(
            str(self.validation_folder) + "/df_performances.csv", index=False
        )


def uncertainty_analysis(
    simulation_name: str,
    simulation_path: str = "Simulation",
    amqp_consumer_adress: Union[str, None] = None,
    sample_size: int = 1000,
    batch_size: int = 100,
    n_jobs: int = -1,
    adx_writer: Union[ADXAndFileWriter, None] = None,
    validation_folder: Union[str, None] = None,
    cold_inputs: dict = {},  # Additional parameters that might be passed to the simulator at each task evaluation
):

    with Timer("[Run Uncertainty Analysis]") as t:

        ua = UncertaintyAnalyzer(
            simulation_name=simulation_name,
            simulation_path=simulation_path,
            amqp_consumer_adress=amqp_consumer_adress,
            sample_size=sample_size,
            batch_size=batch_size,
            n_jobs=n_jobs,
            validation_folder=validation_folder,
            consumers=["Transports", "Stocks", "StocksAtEndOfSimulation", "Performances", "PerformanceAMQP"],
            cold_inputs=cold_inputs,
            timer=t,
        )

        results = ua.execute()

        if adx_writer is not None:
            t.split("Sending stats to ADX")
            adx_writer.write_target_file(
                results["Stock"].to_dict("records"), "StockUncertaintyStatistics"
            )
            adx_writer.write_target_file(
                results["StocksAtEndOfSimulation"].to_dict("records"), "StocksAtEndOfSimulationUncertaintyStatistics"
            )
            adx_writer.write_target_file(
                results["Transport"].to_dict("records"),
                "TransportUncertaintyStatistics",
            )

            t.split("Sent stats to ADX : {time_since_last_split}")

        t.display_message("Running simple simulation to fill ADX")
        # Put back log level to Info for final simulation
        # Reduce log level to Error during optimization
        logger = CosmoEngine.LoggerManager.GetInstance().GetLogger()
        logger.SetLogLevel(logger.eInfo)

        stop_uncertainty = {"Model::@ActivateUncertainties": "false"}

        run_simple_simulation(
            simulation_name=simulation_name,
            simulation_path=simulation_path,
            amqp_consumer_adress=amqp_consumer_adress,
            modifications=stop_uncertainty,
        )
        t.split("Final simulation succeeded : {time_since_last_split}")



parameters_default_values = {
    "Stock": {
        # "id",
        # "Label",
        "PlantName": "",
        "Step": "",
        "PartId": "",
        "MinimalStock": -1,
        "MaximalStock": None,
        # "InitialStock",
        # "InitialValue",
        # "IsInfinite",
        "PurchasingUnitCosts": {},
        "VariablePurchasingUnitCosts": {},
        "CO2UnitEmissions": {},
        "UnitIncomes": {},
        "BacklogWeight": 1,
        "MaximizationWeight": 0,
        "StockPolicy": "None",
        "SourcingPolicy": "Equidistribution",
        "DispatchPolicy": "Equidistribution",
        "ReviewPeriod": 1,
        "FirstReview": 0,
        "Advance": 0,
        "Latitude": 0.0,
        "Longitude": 0.0,
        "IgnoreDownstreamRequiredQuantities": False,
        "SourcingLimit": "Capacity",
        "StorageUnitCosts": {},
        "OrderPoints": {},
        "OrderQuantities": {},
        "OrderUpToLevels": {},
        "SafetyQuantities": {},
        "Demands": {},
        "DemandUncertainties": {},
        "DemandWeights": {},
        "SalesForecasts": {},
    },
    "Transport": {
        # "Label",
        # "source",
        # "target",
        # "Duration",
        "Priority": 0,
        "Mode": "",
        "CustomFees": {},
        "TransportUnitCosts": {},
        "CO2UnitEmissions": {},
        "ActualDurations": {},
        "MinimumOrderQuantities": {},
        "MultipleOrderQuantities": {},
        # "SourcingProportions",
        "TransportUncertaintiesParameter1": {},
        "TransportUncertaintiesParameter2": {},
        "TransportUncertaintiesParameter3": {},
        "TransportUncertaintiesParameter4": {},
        "InitialTransportedQuantities": {},
        "InitialTransportedValues": {},
    },
    "Configuration": {
        "ActivateUncertainties": False,
        "ActivateVariableMachineOpeningRate": False,
        "BatchSize": 0,
        "EmptyObsoleteStocks": False,
        "EnforceProductionPlan": False,
        "FinancialCostOfStock": 0,
        "ManageBacklogQuantities": False,
        "OptimizationObjective": "ServiceLevelMaximization",
        "SimulatedCycles": 1,
        # "StartingDate",
        "StepsPerCycle": 1,
        "TimeStepDuration": 1440,
        "UncertaintiesProbabilityDistribution": "Uniform",
        "IntermediaryStockDispatchPolicy": "DispatchAll",
        "TransportUncertaintiesProbabilityDistribution": "discreteuniform",
        "ActualizeShipments": False,
        "ActivateCorrelatedDemandUncertainties": False,
        "DemandCorrelations": 0,
        "CarbonTax": 0,
        "Kpi": "Profit",
        "OptimizationMode": "maximize",
        "Statistic": "mean",
        "TargetedValue": 80,
        "DecisionVariable": "Order quantity",
        "DecisionVariableMin": 0,
        "DecisionVariableMax": 2000,
        "OptimizationMaximalDuration": 3600,
        "OptimizationAlgorithm": "NGOpt",
        "OptimizationBatchSize": 1,
        "SampleSizeUncertaintyAnalysis": 25,
        "FinalSampleSizeUncertaintyAnalysis": 1000,
        "MaxIterationsForOptim": 250,
        "AutomaticParallelizationConfig": True,
        "MaxNumberOfSimInParallel": 1,
    },
    "contains": {
        # "source",
        # "target",
    },
    "output": {
        # "source",
        # "target",
    },
    "input": {
        # "source",
        # "target",
        "InputQuantity": 1,
    },
    "ProductionResource": {
        # "id",
        # "Label",
        "PlantName": "",
        "ProductionStep": "",
        "ProductionPolicy": "Equidistribution",
        "Latitude": 0.0,
        "Longitude": 0.0,
        "FixedProductionCosts": {},
        # "OpeningTimes",
        "OpeningRates": {},
    },
    "ProductionOperation": {
        # "id",
        # "Label",
        "PlantName": "",
        "IsContractor": False,
        "InvestmentCost": 0,
        "Priority": 0,
        "Duration": 0,
        "QuantitiesToProduce": {},
        "OperatingPerformances": {},
        "CycleTimes": {},
        "RejectRates": {},
        "OperatingPerformanceUncertainties": {},
        "ProductionUnitCosts": {},
        "CO2UnitEmissions": {},
        "MinimumOrderQuantities": {},
        "MultipleOrderQuantities": {},
        # "SourcingProportions",
        # "ProductionProportions",
    },
}
variables_default_values = {
    "Stock": {
        "StorageUnitCosts": 0,
        "OrderPoints": 0,
        "OrderQuantities": 0,
        "OrderUpToLevels": 0,
        "SafetyQuantities": 0,
        "Demands": 0,
        "DemandUncertainties": 0,
        "DemandWeights": 1,
        "SalesForecasts": 0,
        "PurchasingUnitCosts": 0,
        "VariablePurchasingUnitCosts": 0,
        "CO2UnitEmissions": 0,
        "UnitIncomes": 0,
    },
    "Transport": {
        "CustomFees": 0,
        "TransportUnitCosts": 0,
        "CO2UnitEmissions": 0,
        "ActualDurations": 0,
        "MinimumOrderQuantities": 0,
        "MultipleOrderQuantities": 0,
        "TransportUncertaintiesParameter1": 0,
        "TransportUncertaintiesParameter2": 0,
        "TransportUncertaintiesParameter3": 0,
        "TransportUncertaintiesParameter4": 0,
        "InitialTransportedQuantities": 0,
        "InitialTransportedValues": 0,
        "SourcingProportions": 1,
    },
    "Configuration": {},
    "contains": {},
    "output": {},
    "input": {},
    "ProductionResource": {
        "FixedProductionCosts": 0,
        "OpeningTimes": 1440,
        "OpeningRates": 1,
    },
    "ProductionOperation": {
        "QuantitiesToProduce": 0,
        "OperatingPerformances": 1,
        "CycleTimes": 1,
        "RejectRates": 0,
        "OperatingPerformanceUncertainties": 0,
        "ProductionUnitCosts": 0,
        "CO2UnitEmissions": 0,
        "MinimumOrderQuantities": 0,
        "MultipleOrderQuantities": 0,
        "SourcingProportions": 0,
        "ProductionProportions": 0,
    },
}



import pandas
from typing import Union
import comets as co
import numpy as np

from Supplychain.Wrappers.simulator import CosmoEngine
from Supplychain.Generic.timer import Timer
from Supplychain.Run.uncertainty_analysis_comets import UncertaintyAnalyzer
from Supplychain.Wrappers.environment_variables import EnvironmentVariables


class StochasticOptimizer:
    """
    Object in charge of performing an optimization on the results of an uncertainty analysis.
    It is possible to run an optimization without uncertainties if the dataset has ActivateUncertainties set to false.
    The main method of the StochasticOptimizer is run, which returns the results of the optimization.

    Args:
        simulation_name (str): Name of simulation, used by the probes
        KPI (str): Which KPI is objective of the optimization. Choose among "Profit", OPEX, AverageStockValue, ServiceLevelIndicator, ServiceLevelSatisfaction, CO2Emissions, TotalServedQuantity, IndividualServiceLevelSatisfaction.
        stat (str, optional): Which statistics of the KPI we care about. Choose among mean, std,	sem,	quantile 5%,	quantile 10%,	... ,	quantile 95%. Defaults to "mean".
        optimization_mode (str, optional): Optimization mode, choose among "maximize"/"minimize"/"target". Defaults to "maximize".
        target_value (float, optional): Target value to reach if optimization_mode is "target". Defaults to 0. Note service levels are expressed between 0 and 100.
        service_level_of_stocks (list, optional): If KPI is IndividualServiceLevelSatisfaction, choose for which stocks we want to optimize the service level. Defaults to [], which means for all stocks.
        decision_variables_entity_type (str, optional): Entity type for which we have levers (decision variables) we want to change. Defaults to "Stock".
        decision_variables_entities (list, optional): Choose which particular entities have levers we want to change, defaults to all entities of type above. Defaults to [].
        decision_variable_attribute (str, optional): Choose attribute which is a decision variable (lever) for the optimization. Defaults to "OrderQuantitySchedule".
        decision_variable_attribute_schedulable (bool, optional): Whether the chosen attribute is schedulable. Defaults to True.
        decision_variable_min (float, optional): Minimum value that the attribute can take (same for all entities), defines search space for the optimization. Defaults to 0.
        decision_variable_max (float, optional): Maximum value that the attribute can take (same for all entities), defines search space for the optimization. Defaults to 100.
        simulation_path (str): Name of the simulation file. Defaults to "Simulation".
        amqp_consumer_adress (Union[str, None], optional): Adress of consumer to send probe results to. Defaults to None.
        sample_size_uncertainty_analysis (int, optional): Number of simulations runs by the uncertainty analysis. Defaults to 100.
        batch_size_uncertainty_analysis (int, optional): Number of simulations runs that are run in a same batch by the uncertainty analysis. Defaults to 100.
        max_iterations_for_optim (int, optional): Max number of run of an uncertainty analysis during the optimization. Defaults to 100.
        max_duration_of_optim (int, optional): Max duration (s) of the optim, the optim runs until either this duration, or max_iterations_for_optim is reached. Defaults to 3600*10 (10h).
        optimization_algorithm (str, optional): Name of optimization algorithm. Defaults to NGOpt.
        optim_batch_size (int, optional): Choose batch size of optimization (highly dependent of algorithm choice). Defaults to 1.
        n_jobs (int, optional): Choose number of cpus used in parallel by the optimization. Note that the uncertainty analysis is also parallelized independently. Defaults to 1. Use -1 for all cpus available.
    """

    def __init__(
        self,
        simulation_name: str,
        KPI: str = "Profit",
        stat: str = "mean",
        optimization_mode: str = "maximize",
        target_value: float = 0,
        service_level_of_stocks: list = [],
        decision_variables_entity_type="Stock",
        decision_variables_entities: list = [],
        decision_variable_attribute: str = "SafetyQuantitySchedule",
        decision_variable_attribute_schedulable: bool = True,
        decision_variable_min: float = 0,
        decision_variable_max: float = 100,
        simulation_path: str = "Simulation",
        amqp_consumer_adress: Union[str, None] = None,
        sample_size_uncertainty_analysis: int = 25,
        batch_size_uncertainty_analysis: int = 25,
        max_iterations_for_optim: int = 1000,
        max_duration_of_optim: int = 3600 * 10,  # 10h
        optimization_algorithm: str = "NGOpt",
        optim_batch_size: int = 1,
        n_jobs: int = 1,
        n_jobs_ua: int = -1,
    ):
        self.simulation_name = simulation_name
        self.KPI = KPI
        self.stat = stat
        self.optimization_mode = optimization_mode
        self.target_value = target_value
        self.service_level_of_stocks = service_level_of_stocks
        if isinstance(decision_variables_entity_type, str):
            decision_variables_entity_type = [decision_variables_entity_type]
        self.decision_variables_entity_type = decision_variables_entity_type
        self.decision_variables_entities = decision_variables_entities
        self.decision_variable_attribute = decision_variable_attribute
        self.decision_variable_attribute_schedulable = (
            decision_variable_attribute_schedulable
        )
        self.decision_variable_min = decision_variable_min
        self.decision_variable_max = decision_variable_max
        self.simulation_path = simulation_path
        self.amqp_consumer_adress = amqp_consumer_adress
        self.sample_size_uncertainty_analysis = sample_size_uncertainty_analysis
        self.batch_size_uncertainty_analysis = batch_size_uncertainty_analysis
        self.max_iterations_for_optim = max_iterations_for_optim
        self.max_duration_of_optim = max_duration_of_optim
        self.optimization_algorithm = optimization_algorithm
        self.optim_batch_size = optim_batch_size
        self.n_jobs_optim = n_jobs
        self.n_jobs_ua = n_jobs_ua

    def initialize(self):
        """
        Initialize the optimization by collecting information from the simulator, creating the optimization space and task.
        """
        # Collect simulator information
        cosmo_interface = co.CosmoInterface(
            self.simulation_path, custom_sim_engine=CosmoEngine
        )
        cosmo_interface.initialize()

        if self.decision_variables_entities == []:
            entities = []
            for entity_type in self.decision_variables_entity_type:
                entities.append(self._get_entities(cosmo_interface, entity_type))

        else:
            entities = [self.decision_variables_entities]
        datapaths = []
        for entities_grouped_by_type in entities:
            datapaths += self._get_list_of_datapaths(
                cosmo_interface,
                entities_grouped_by_type,
                self.decision_variable_attribute,
            )

        list_of_init_values = self._get_init_value(
            cosmo_interface,
            datapaths,
            (self.decision_variable_max - self.decision_variable_min) / 2,
        )
        self.ActivateUncertainties = cosmo_interface.get_outputs(
            ["Model::@ActivateUncertainties"]
        )["Model::@ActivateUncertainties"]

        list_of_types = self._get_type(datapaths)
        cosmo_interface.terminate()

        # If uncertainties are not activated in the model, set standard parameters of uncertainty analysis
        if not self.ActivateUncertainties:
            self.sample_size_uncertainty_analysis = 1
            self.batch_size_uncertainty_analysis = 1
            self.stat = "mean"

        # Create decision variable space
        self.space = self._create_optim_space(
            datapaths,
            list_of_init_values,
            list_of_types,
            self.decision_variable_min,
            self.decision_variable_max,
        )

        # Create uncertainty analyzer
        consumers = ["Performances"]
        if self.KPI == "IndividualServiceLevelSatisfaction":
            consumers.append("StocksAtEndOfSimulation")

        self.ua = UncertaintyAnalyzer(
            simulation_name=self.simulation_name,
            simulation_path=self.simulation_path,
            amqp_consumer_adress=None,
            sample_size=self.sample_size_uncertainty_analysis,
            batch_size=self.batch_size_uncertainty_analysis,
            consumers=consumers,
            cold_inputs={},
            validation_folder=None,
            n_jobs=self.n_jobs_ua,
        )

        # Initialize uncertainty analysis
        self.ua.initialize_simulator_interface()
        self.ua.collect_simulation_parameters()
        self.ua.create_encoder()
        self.ua.create_get_outcomes()
        self.ua.create_sampling()

        # Declare task running the uncertainty analysis
        def task(
            input_parameter_set,
            ua=self.ua,
            activateuncertainties=self.ActivateUncertainties,
            decision_variable_attribute_schedulable=self.decision_variable_attribute_schedulable,
            _compute_objective=_compute_objective,
            KPI=self.KPI,
            stat=self.stat,
            optimization_mode=self.optimization_mode,
            target_value=self.target_value,
            service_level_of_stocks=self.service_level_of_stocks,
        ):
            if activateuncertainties:
                np.random.seed()
            else:
                np.random.seed(0)

            new_input_parameter_set = {}
            if decision_variable_attribute_schedulable:
                for key, value in input_parameter_set.items():
                    new_input_parameter_set[key] = {0: value}  # Transform to schedule
            else:
                new_input_parameter_set = input_parameter_set

            ua.create_task(new_input_parameter_set)
            ua.run_experiment()
            ua.reformat_results()

            objective = _compute_objective(
                ua.results,
                KPI=KPI,
                stat=stat,
                optimization_mode=optimization_mode,
                target_value=target_value,
                service_level_of_stocks=service_level_of_stocks,
            )
            return objective

        self.task = task

    def run(self):
        """Run the optimization

        Returns:
            Tuple: (kpi_results, optimal_decision_variables, optimal_decision_variables_df, optimization_history)
                kpi_results is a dictionary containing key Objective (the optimal value of the objective function) and kay KPI (the optimal value of the KPI, may be different from the objective)
                optimal_decision_variables is a dictionary (ParameterSet) with keys that are datapaths of decision variables, and values that are their recommended value by the optimization.
                    It may be used to launch a new uncertainty analysis with the optimal decision variables.
                optimal_decision_variables_df is a pandas dataframe containing the recommended choice of decision variables, with columns  Datapath, Value, Attribute, Entity.
                optimization_history is a pandas dataframe containing the optimization metrics, with columns ObjectiveValue, KPIValue, Iteration, KPI, Stat, optimization_mode, target_value.
        """
        with Timer("[Run stochastic optimization]") as t:
            self.initialize()

            opt = co.Optimization(
                space=self.space,
                task=self.task,
                objective="Objective",
                maximize=(self.optimization_mode == "maximize"),
                algorithm=self.optimization_algorithm,
                batch_size=self.optim_batch_size,
                stop_criteria={
                    "max_evaluations": self.max_iterations_for_optim,
                    "max_duration": self.max_duration_of_optim,
                },
                n_jobs=self.n_jobs_optim,
                save_optimization_history=True,
            )

            opt.run()

            kpi_results = self._reformat_kpi_results(opt)
            optimization_history = self._reformat_optimization_history(
                opt.optimization_history
            )
            optimal_decision_variables = opt.results["Optimal variables"]
            optimal_decision_variables_df = self._reformat_decision_variables_results(
                optimal_decision_variables
            )

            if self.optimization_algorithm == "NGOpt" and self.optim_batch_size == 1:
                del opt.optimizationalgorithm

            t.display_message("Running simple simulation to fill ADX")
            # Put back log level to Info for final simulation
            # Reduce log level to Error during optimization
            logger = CosmoEngine.LoggerManager.GetInstance().GetLogger()
            logger.SetLogLevel(logger.eInfo)

        return (
            kpi_results,
            optimal_decision_variables,
            optimal_decision_variables_df,
            optimization_history,
        )

    def _create_optim_space(
        self, datapaths, list_of_init_values, list_of_types, minimum, maximum
    ):
        space = []
        for datapath, init, type in zip(datapaths, list_of_init_values, list_of_types):
            space.append(
                {
                    "name": datapath,
                    "type": str(type),
                    "bounds": [minimum, maximum],
                    "init": init,
                }
            )
        return space

    def _get_init_value(self, cosmo_interface, list_of_datapaths, default):
        list_of_init_values = []
        inits = cosmo_interface.get_outputs(list_of_datapaths)
        for datapath in list_of_datapaths:
            if isinstance(inits[datapath], dict):
                if 0 not in inits[datapath]:
                    list_of_init_values.append(default)
                else:
                    if (
                        inits[datapath][0] <= self.decision_variable_max
                        and inits[datapath][0] >= self.decision_variable_min
                    ):
                        list_of_init_values.append(inits[datapath][0])
                    else:
                        list_of_init_values.append(default)
            else:
                if (
                    inits[datapath] <= self.decision_variable_max
                    and inits[datapath] >= self.decision_variable_min
                ):
                    list_of_init_values.append(inits[datapath])
                else:
                    list_of_init_values.append(default)
        return list_of_init_values

    def _get_list_of_datapaths(self, cosmo_interface, list_of_entities, attribute):
        # Retrieve the datapath of attributes for all entities in list_of_entities
        list_of_datapaths = []

        datapaths = cosmo_interface.get_datapaths()

        for entity in list_of_entities:
            list_of_datapaths.append(
                list(
                    filter(
                        lambda x: str("{Entity}" + entity + f"::@{attribute}") in x,
                        datapaths,
                    )
                )[0]
            )

        return list_of_datapaths

    def _get_entities(self, cosmo_interface, decision_variables_entity_type):
        # Retrieve the name of all the entities whose type corresponds to the one selected by the user

        list_of_entities = sorted(
            entity.GetName()
            for entity in cosmo_interface.sim.GetModel().FindEntitiesByType(
                decision_variables_entity_type
            )
        )

        if list_of_entities == []:
            raise ValueError(
                f"No entity of type {decision_variables_entity_type} is present in the instance, the optimization is not supported."
            )

        return list_of_entities

    def _reformat_decision_variables_results(self, decision_variables):
        results = pandas.DataFrame(
            [
                {"Datapath": key, "Value": value}
                for (key, value) in decision_variables.items()
            ]
        )
        results["Attribute"] = self.decision_variable_attribute
        results["Entity"] = results["Datapath"].apply(
            lambda x: x.split("{Entity}")[-1].split("::@")[0]
        )
        results["SimulationRun"] = EnvironmentVariables.simulation_id
        return results

    def _reformat_optimization_history(self, list_of_results):
        rows = []
        if self.KPI != "IndividualServiceLevelSatisfaction":
            for ((iteration, objective), kpivalue) in zip(
                enumerate(list_of_results["mean_objective"]),
                list_of_results["mean_task_outputs"],
            ):
                rows.append(
                    {
                        "ObjectiveValue": objective,
                        "KPIValue": kpivalue["KPI"],
                        "Iteration": iteration,
                        "KPI": self.KPI,
                        "Stat": self.stat,
                        "optimization_mode": self.optimization_mode,
                        "target_value": self.target_value,
                        "SimulationRun": EnvironmentVariables.simulation_id,
                    }
                )
        else:
            for ((iteration, objective), kpivalue) in zip(
                enumerate(list_of_results["mean_objective"]),
                list_of_results["all_task_outputs"],
            ):
                rows.append(
                    {
                        "ObjectiveValue": objective,
                        "KPIValue": np.mean(kpivalue[0]["KPI"]),
                        "Iteration": iteration,
                        "KPI": self.KPI,
                        "Stat": self.stat,
                        "optimization_mode": self.optimization_mode,
                        "target_value": self.target_value,
                        "SimulationRun": EnvironmentVariables.simulation_id,
                    }
                )
        return pandas.DataFrame(rows)

    def _reformat_kpi_results(self, opt):
        results = self.task(opt.results["Optimal variables"])
        results["SimulationRun"] = EnvironmentVariables.simulation_id
        return pandas.DataFrame(results, index=[0])

    def _get_type(self, list_of_datapaths):
        """
        Returns the type of each of the decision variable
        """
        type_mapping = {
            "OpeningTimeSchedule": "float",
            "OrderPointSchedule": "float",
            "OrderQuantitySchedule": "float",
            "SafetyQuantitySchedule": "float",
            "SourcingProportionSchedule": "float",
            "ReviewPeriod": "int",
            "Advance": "int",
        }
        list_of_types = []
        for datapaths in list_of_datapaths:
            list_of_types.append(type_mapping[self.decision_variable_attribute])
        return list_of_types


# Implementing a function instead of a method to be able to use it in the multiprocessing module
def _compute_objective(
    results, KPI, stat, optimization_mode, target_value, service_level_of_stocks
):
    if KPI != "IndividualServiceLevelSatisfaction":
        results["Performances"] = results["Performances"].set_index("KPI")
        kpi = results["Performances"].loc[KPI, stat]
        if optimization_mode == "target":
            objective = {"Objective": (kpi - target_value) ** 2, "KPI": kpi}
        else:
            objective = {"Objective": kpi, "KPI": kpi}
    else:
        df_service = results["StocksAtEndOfSimulation"][
            results["StocksAtEndOfSimulation"]["Indicator"] == "ServiceLevelSatisfaction"
        ]
        if service_level_of_stocks != []:
            df_service = df_service[df_service["id"].isin(service_level_of_stocks)]
        vector_of_service_levels = df_service[stat].to_numpy()
        if optimization_mode == "target":
            distance_to_target = (vector_of_service_levels - target_value) ** 2
            objective = {
                "Objective": np.max(distance_to_target),  # np.sum(distance_to_target)
                "KPI": list(vector_of_service_levels),
            }
        else:
            objective = {
                "Objective": np.sum(vector_of_service_levels),
                "KPI": list(vector_of_service_levels),
            }
    return objective



class StockConsumer:
    """
    Python Consumer for stocks indicators at each time step.
    """

    def __init__(self):
        self.memory = list()

    def Consume(self, p_data):
        probe_output = self.engine.StocksProbeOutput.Cast(p_data)
        f = probe_output.GetFacts()
        timestep = int(probe_output.GetProbeRunDimension().GetProbeOutputCounter())
        for data in f:
            fact = [
                str(data.GetAttributeAsString("id")),
                timestep,
                float(data.GetAttributeAsDouble("Demand")),
                float(data.GetAttributeAsDouble("RemainingQuantity")),
                float(data.GetAttributeAsDouble("ServedQuantity")),
                float(data.GetAttributeAsDouble("UnservedQuantity")),
                float(data.GetAttributeAsDouble("ServiceLevel")),
                float(data.GetAttributeAsDouble("Value")),
            ]
            self.memory.append(fact)


class PerformanceConsumer:
    """
    Python Consumer for performance indicators at the end of the simulation.
    """

    def __init__(self):
        self.memory = list()

    def Consume(self, p_data):
        probe_output = self.engine.PerformanceIndicatorsProbeOutput.Cast(p_data)
        f = probe_output.GetFacts()
        for data in f:
            fact = {
                "OPEX": float(data.GetAttributeAsDouble("OPEX")),
                "Profit": float(data.GetAttributeAsDouble("Profit")),
                "AverageStockValue": float(
                    data.GetAttributeAsDouble("AverageStockValue")
                ),
                "ServiceLevelIndicator": float(
                    data.GetAttributeAsDouble("ServiceLevelIndicator")
                ),
                "CO2Emissions": float(data.GetAttributeAsDouble("CO2Emissions")),
                "TotalDemand": float(data.GetAttributeAsDouble("TotalDemand")),
                "TotalServedQuantity": float(
                    data.GetAttributeAsDouble("TotalServedQuantity")
                ),
                "ServiceLevelSatisfaction": data.GetTotalServedQuantity().GetAsFloat()
                / data.GetTotalDemand().GetAsFloat()
                * 100
                if data.GetTotalDemand().GetAsFloat() != 0
                else 0,
            }
            self.memory.append(fact)


class StocksAtEndOfSimulationConsumer:
    """
    Python Consumer for stocks global indicators at the end of the simulation.
    """

    def __init__(self):
        self.memory = list()

    def Consume(self, p_data):
        probe_output = self.engine.StocksAtEndOfSimulationProbeOutput.Cast(p_data)
        f = probe_output.GetFacts()
        for data in f:
            fact = {
                "id": str(data.GetAttributeAsString("id")),
                "TotalDemand": float(data.GetAttributeAsDouble("TotalDemand")),
                "TotalServedQuantity": float(data.GetAttributeAsDouble("TotalServedQuantity")),
                "ServiceLevelSatisfaction": 100 * float(data.GetAttributeAsDouble("ServiceLevelSatisfaction")),
                "CycleServiceLevel": float(data.GetAttributeAsDouble("CycleServiceLevel")),
            }
            self.memory.append(fact)



from copy import deepcopy
import comets as co
from Supplychain.Wrappers.simulator import CosmoEngine
from Supplychain.Wrappers.environment_variables import EnvironmentVariables

"""
-------------------------------------------------
-------------------------------------------------
-------------------------------------------------
Helper functions used in the uncertainty analysis
-------------------------------------------------
-------------------------------------------------
-------------------------------------------------
"""


"""
-------------------------------------------------
Simple helper functions
-------------------------------------------------
"""


def extend_simple_dic(my_dic, number_of_iterations):
    """Function to extend dictionaries of schedulable attributes.

    Args:
        my_dic (dict): dictionary of scheduled values such as {0: 3, 6: 4, 7: 3, 8: 2, 9: 8, 10: 40}
        number_of_iterations (int): total number of time steps of the schedule

    Returns:
        dict: extended dictionary for all time steps {0: 3, 1: 3, 2: 3, 3: 3, 4: 3, 5: 3, 6: 4, 7: 3, 8: 2, 9: 8, 10: 40}
    """
    if my_dic != {}:  # checking that the dic isn't empty
        extended_dic = {
            0: my_dic[0]
        }  # We assume that the uncertainty starts at the first time step
        for i in range(1, number_of_iterations):
            if i in my_dic:
                extended_dic[i] = deepcopy(my_dic[i])
            else:
                extended_dic[i] = deepcopy(extended_dic[i - 1])
    else:
        extended_dic = {}
    return extended_dic


def extend_dic(my_dic, number_of_iterations):
    """Function to extend dictionaries of schedulable attributes.

    Args:
        my_dic (dict): dictionary of dictionary of scheduled values
        number_of_iterations (int): total number of time steps of the schedule

    Returns:
        dict: dictionary containing extended dictionary for all time steps
    """
    extended_dic = {}
    for entity in my_dic.keys():
        extended_dic[entity] = extend_simple_dic(my_dic[entity], number_of_iterations)
    return extended_dic


def add_tag(tag, name):
    return tag + name


def add_tag_to_parameterset(tag, parameterset):
    """Add string tag in front of the keys of a parameterset"""
    return {add_tag(tag, key): value for key, value in parameterset.items()}


def check_distrib_parameters(my_distribution, param1, param2, param3, param4):
    """Transform distribution parameters that don't match CoMETS format

    Args:
        my_distribution (str): name of distribution
        param1 (float): dictionary of values for first parameter
        param2 (float): dictionary of values for second parameter
        param3 (float): dictionary of values for third parameter
        param4 (float): dictionary of values for fourth parameter

    Returns:
        list: list containing dictionaries with modified values for param1, 2, 3 and 4
    """
    # In CoMETS the upper bound of the discreteuniform distribution is excluded.
    # However, in supply chain it is included. Therefore, we need to add 1 to the upper bound
    # so it matches CoMETS
    if my_distribution == "discreteuniform":
        for entity in param2.keys():
            for keys in param2[entity].keys():
                param2[entity][keys] += 1

    # In CoMETS the two arguments of the uniform distribution are loc and scale
    # and the interval of the distribution is the following:  [loc, loc+scale]
    # However, in supply chain the two parameters of the uniform distribution are
    # [lower, upper]. Therefor, upper needs to be mapped to scale
    if my_distribution == "uniform":
        for entity in param2.keys():
            for keys in param2[entity].keys():
                param2[entity][keys] = param2[entity][keys] - param1[entity][keys]

    if my_distribution == "betabinom":
        for entity in param1.keys():
            for keys in param1[entity].keys():
                param1[entity][keys] = round(param1[entity][keys])

    if my_distribution == "binomial":
        for entity in param1.keys():
            for keys in param1[entity].keys():
                param1[entity][keys] = round(param1[entity][keys])

    if my_distribution == "hypergeom":
        for entity in param1.keys():
            for keys in param1[entity].keys():
                param1[entity][keys] = round(param1[entity][keys])
                param2[entity][keys] = round(param2[entity][keys])
                param3[entity][keys] = round(param3[entity][keys])

    return [param1, param2, param3, param4]


"""
-------------------------------------------------
Functions that collect simulator attributes before the analysis
-------------------------------------------------
"""


def get_transports(cosmo_interface):
    """Function to get the list of all the transports in the simulation"""
    transports_list = []
    transports = cosmo_interface.sim.get_entities_names_by_type(
        entity_type="TransportOperation"
    )
    for keys in transports:
        transports_list.append(keys)
    return transports_list


def get_stocks(cosmo_interface):
    """Function to get the list of all the stocks that have uncertain demand"""
    uncertain_stocks = []
    for stock in cosmo_interface.sim.get_entities_by_type("Stock"):
        demands = CosmoEngine.DataTypeMapInterface.Cast(stock.GetAttribute("Demand"))
        stock_name = stock.GetName()
        for time_step in demands.GetKeys():
            demand = demands.GetAt(time_step)
            if demand.GetAttribute("DemandRelativeUncertainty").Get() > 0:
                uncertain_stocks.append(stock_name)
                break
    return uncertain_stocks


def get_attribute(cosmo_interface, attribute):
    """Get value of attribute in the model"""
    return cosmo_interface.get_outputs([attribute])[attribute]


def get_max_time_step(cosmo_interface):
    time_step_per_cycle = get_attribute(cosmo_interface, "Model::@TimeStepPerCycle")
    number_of_cycle = get_attribute(cosmo_interface, "Model::@NumberOfCycle")
    max_time_step = time_step_per_cycle * number_of_cycle
    return max_time_step


def collect_transport_information(cosmo_interface, transports_names):
    """Collect information regarding transports in the model:
    which transports have uncertain duration,
    what are the theoretical durations,
    what are the parameters of the probability distribution of transport duration for each transport.
    """
    list_of_transports = transports_names.copy()
    actual_duration_schedule = {}
    uncertainty_param_1 = {}
    uncertainty_param_2 = {}
    uncertainty_param_3 = {}
    uncertainty_param_4 = {}
    for transport in transports_names:

        # If their is no ActualDuration in the TransportSchedules column of the dataset,
        # we use by default the attribute duration, in the Transport column of the dataset.
        if (
            cosmo_interface.get_outputs(
                [
                    f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@ActualDurationSchedule"
                ]
            )[
                f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@ActualDurationSchedule"
            ]
            == {}
        ):
            duration = cosmo_interface.get_outputs(
                [
                    f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@Duration"
                ]
            )[f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@Duration"]

            # Use a dict format so that the function "extended_dict" defined below can be applied
            actual_duration_schedule[transport] = {0: duration}
        else:
            actual_duration_schedule[transport] = cosmo_interface.get_outputs(
                [
                    f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@ActualDurationSchedule"
                ]
            )[
                f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@ActualDurationSchedule"
            ]

        uncertainty_param_1[transport] = cosmo_interface.get_outputs(
            [
                f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@TransportUncertaintiesParameter1"
            ]
        )[
            f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@TransportUncertaintiesParameter1"
        ]
        uncertainty_param_2[transport] = cosmo_interface.get_outputs(
            [
                f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@TransportUncertaintiesParameter2"
            ]
        )[
            f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@TransportUncertaintiesParameter2"
        ]
        uncertainty_param_3[transport] = cosmo_interface.get_outputs(
            [
                f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@TransportUncertaintiesParameter3"
            ]
        )[
            f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@TransportUncertaintiesParameter3"
        ]
        uncertainty_param_4[transport] = cosmo_interface.get_outputs(
            [
                f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@TransportUncertaintiesParameter4"
            ]
        )[
            f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@TransportUncertaintiesParameter4"
        ]
        # If the transport has no parameters, its transport duration will not be part of the uncertainty analysis
        if (
            uncertainty_param_1[transport] == uncertainty_param_2[transport]
            and uncertainty_param_1[transport] == {}
        ):
            list_of_transports.remove(transport)

    return (
        list_of_transports,
        actual_duration_schedule,
        uncertainty_param_1,
        uncertainty_param_2,
        uncertainty_param_3,
        uncertainty_param_4,
    )


"""
-------------------------------------------------
Functions that collect simulator outputs
-------------------------------------------------
"""


def collect_simulated_transport_output(transports_names, modelinterface, max_time_step):
    """
    Function that returns a parameterset with the transport duration for each transport at the end of the simulation
    The transport duration is separated for each time step. The output parameterset (for a simulation
    with 1 TransportOperation: U) will have the following format:
    {Model[...]U::@ActualDurationSchedule__&@&__0': 10,
           [...],
      Model[...]U::@ActualDurationSchedule__&@&__10': 7}
    """
    transport_duration = {}
    transport_duration_transformed = {}
    for transport in transports_names:
        if (
            modelinterface.get_outputs(
                [
                    f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@ActualDurationSchedule"
                ]
            )[
                f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@ActualDurationSchedule"
            ]
            == {}
        ):
            duration = modelinterface.get_outputs(
                [
                    f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@Duration"
                ]
            )[f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@Duration"]
            # Use a dict format so that the function "extend_dict" can be applied
            actual_duration_schedule = {0: duration}
        else:
            actual_duration_schedule = modelinterface.get_outputs(
                [
                    f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@ActualDurationSchedule"
                ]
            )[
                f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@ActualDurationSchedule"
            ]
        transport_duration[
            f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@ActualDurationSchedule"
        ] = extend_simple_dic(
            actual_duration_schedule,
            max_time_step,
        )
        time_step = 0
        for value in transport_duration[
            f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@ActualDurationSchedule"
        ].values():
            transport_duration_transformed[f"{transport}__&@&__ActualDuration__&@&__{time_step}"] = value
            time_step += 1
    return transport_duration_transformed


def collect_simulated_stock_output(consumer_memory):
    """
    This function transforms the consumer memory from a list of list to a list of
    ParameterSet. Note that each sublist in the initial format is transformed into
    a parameterset of 6 parameters : Demand, RemainingQuantity, ServedQuantity, UnservedQuantity, ServiceLevel, Value
    """
    dic_of_parameterset = {}
    for elements in consumer_memory:
        dic_of_parameterset[
            str(elements[0] + "__&@&__" + "Demand" + "__&@&__" + str(elements[1]))
        ] = elements[2]
        dic_of_parameterset[
            str(
                elements[0]
                + "__&@&__"
                + "RemainingQuantity"
                + "__&@&__"
                + str(elements[1])
            )
        ] = elements[3]
        dic_of_parameterset[
            str(
                elements[0]
                + "__&@&__"
                + "ServedQuantity"
                + "__&@&__"
                + str(elements[1])
            )
        ] = elements[4]
        dic_of_parameterset[
            str(
                elements[0]
                + "__&@&__"
                + "UnservedQuantity"
                + "__&@&__"
                + str(elements[1])
            )
        ] = elements[5]
        dic_of_parameterset[
            str(elements[0] + "__&@&__" + "ServiceLevel" + "__&@&__" + str(elements[1]))
        ] = elements[6]
        dic_of_parameterset[
            str(elements[0] + "__&@&__" + "Value" + "__&@&__" + str(elements[1]))
        ] = elements[7]

    return dic_of_parameterset


def collect_simulated_stock_final_output(consumer):
    """
    This function transforms the consumer memory to a dictionary of
    ParameterSets, where one ParameterSet is used for each stock-category.
    """
    categories = ("TotalDemand", "TotalServedQuantity", "ServiceLevelSatisfaction", "CycleServiceLevel")
    return {
        f"{elements['id']}__&@&__{category}": elements[category]
        for elements in consumer
        for category in categories
    }


"""
-------------------------------------------------
Functions that create the uncertainty analysis "sampling"
-------------------------------------------------
"""


def create_transport_distribution_sampling(
    distribution,
    param1,
    param2,
    param3,
    param4,
    transports,
    distribution_parameters,
    number_of_time_steps,
):
    """
    Function to create the uncertainty analysis sampling for the transport durations according to CoMETS format
    This function will create one variable for each time step of each transports

    Args:
        distribution (str): name of the probability distribution used
        param1 (float): values by transport by time step for the first parameter of the distribution
        param2 (float): values by transport by time step for the second parameter of the distribution
        param3 (float): values by transport by time step for the third parameter of the distribution
        param4 (float): values by transport by time step for the fourth parameter of the distribution
        transports (list):list of all the uncertain transports names
        distribution_parameters (str): list of the parameters names required by CoMETS sampler for the given distribution
        number_of_time_steps (int): number of simulated time steps


    Returns:
        list: list of sampling variables according to CoMETS format
    """
    sampling = []
    all_parameters = check_distrib_parameters(
        distribution, param1, param2, param3, param4
    )
    for transport in transports:
        considered_parameters = [
            (position, parameter)
            for position, parameter in enumerate(distribution_parameters)
            if all_parameters[position][transport]
        ]
        for t in range(number_of_time_steps):
            parameters = {
                parameter: all_parameters[position][transport][t]
                for position, parameter in considered_parameters
            }
            sampling.append(
                {
                    "name": f"{transport}_@_{t}",
                    "sampling": distribution,
                    "parameters": parameters,
                }
            )
    return sampling


def create_demand_generator(extended_demands, number_of_time_steps, DemandCorrelations):
    """Function to create the uncertainty analysis sampling for the demands according to CoMETS format.
    Defines a generator of uncertain time series for each stock.

    Args:
        extended_demands (dict): Demand attributes obtained from the simulator
        number_of_time_steps (int): _description_
        DemandCorrelations (float): amount of correlation between consecutive time steps

    Returns:
        list: list of sampling variables according to CoMETS format
    """
    sampling = []
    for stock, demand_attribute in extended_demands.items():
        mean_demand = []
        uncertainties = []
        for t in range(number_of_time_steps):

            demand = demand_attribute[t]["ExternalDemand"]
            mean_demand.append(demand)
            uncertainties.append(
                demand * demand_attribute[t]["DemandRelativeUncertainty"]
            )  # Uncertainty proportional to demand, DemandRelativeUncertainty*Demand is the standard deviation
        sampling.append(
            {
                "name": f"{stock}",
                "sampling": co.TimeSeriesSampler(
                    correlation=DemandCorrelations,
                    dimension=number_of_time_steps,
                    forecast=mean_demand,
                    uncertainties=uncertainties,
                    minimum=0,
                ),
            }
        )
    return sampling


"""
-------------------------------------------------
Function that reformat the outputs of the uncertainty analysis
-------------------------------------------------
"""


def transform_data(data, timestep=True):
    """
    Transform output data so that it matches the ADX table format
    """
    df = data.copy()
    index_cols = [
        "id",
        "Category",
        "TimeStep"
    ]
    if not timestep:
        index_cols.remove("TimeStep")
    if not df.empty:
        df.loc[:, "SimulationRun"] = EnvironmentVariables.simulation_id
        df[index_cols] = df["index"].str.split(pat="__&@&__", expand=True)
    else:
        index_cols.insert(0, "SimulationRun")
        df[index_cols] = None
    cols = [
        "SimulationRun",
        "TimeStep",
        "id",
        "Category",
        "quantile 5%",
        "quantile 25%",
        "quantile 50%",
        "quantile 75%",
        "quantile 95%",
        "mean",
        "sem",
    ]
    if not timestep:
        cols.remove("TimeStep")
    df = df[cols]
    df.rename(
        columns={
            "quantile 5%": "FirstVigintile",
            "quantile 25%": "FirstQuartile",
            "quantile 50%": "Median",
            "quantile 75%": "LastQuartile",
            "quantile 95%": "LastVigintile",
            "mean": "Mean",
            "sem": "SE",
        },
        inplace=True,
    )
    return df



import csv
import glob
import io
import json
import os
import sys

from collections import defaultdict

from Supplychain.Generic.folder_io import FolderReader


class CSVReader(FolderReader):
    extension = ".csv"

    def refresh(self):
        filenames = glob.glob(os.path.join(self.input_folder, "*" + self.extension))
        self.files = defaultdict(list)
        csv.field_size_limit(sys.maxsize)

        for filename in filenames:
            with open(filename, "r") as f:
                # Read every file in the input folder
                current_filename = os.path.basename(filename)[:-len(self.extension)]
                too_many = set()
                too_few = set()
                reader = csv.DictReader(f)
                try:
                    for row in reader:
                        new_row = dict()
                        for key, value in row.items():
                            if key is None:
                                too_many.add(str(reader.line_num))
                                continue
                            if value is None:
                                too_few.add(str(reader.line_num))
                                continue
                            if key in self.id_keys:
                                new_row[key] = None if value is None else str(value)
                                continue
                            try:
                                # Try to convert any json row to dict object
                                converted_value = json.load(io.StringIO(value))
                            except json.decoder.JSONDecodeError:
                                converted_value = value
                            if converted_value == '':
                                converted_value = None
                            if converted_value is not None or self.keep_nones:
                                new_row[key] = converted_value
                        self.files[current_filename].append(new_row)
                except csv.Error as e:
                    self.errors[current_filename] = f"Error while reading {filename}, at line {reader.line_num}: {e}"
                    continue
                if too_many or too_few:
                    error = f"Error while reading {filename}.\nHeader fields: {', '.join(reader.fieldnames)}."
                    if too_many:
                        lines = (str(line) for line in sorted(too_many))
                        error += f"\nMore fields than header in line{'s' if len(too_many) > 1 else ''} {', '.join(lines)}."
                    if too_few:
                        lines = (str(line) for line in sorted(too_few))
                        error += f"\nLess fields than header in line{'s' if len(too_few) > 1 else ''} {', '.join(lines)}."
                    self.errors[current_filename] = error

        if self.errors and self.raise_errors:
            message = "\n".join(self.errors.values())
            raise Exception(message)

    def __init__(self,
                 input_folder: str = "Input",
                 keep_nones: bool = True,
                 raise_errors: bool = True):

        FolderReader.__init__(self,
                              input_folder=input_folder,
                              keep_nones=keep_nones)
        self.raise_errors = raise_errors
        self.errors = {}

        self.refresh()
